"""Implements :class:`bidict.frozenbidict`."""

from .compat import viewitems
from ._common import BidirectionalMapping, _missing
from collections import Hashable


class frozenbidict(BidirectionalMapping, Hashable):
    """Immutable, hashable bidict type."""

    def __hash__(self):
        """Return the hash of this bidict."""
        h = getattr(self, '_hash', _missing)
        if h is not _missing:
            return h
        itemsview = viewitems(self)
        if hasattr(itemsview, '_hash'):  # e.g. collections.ItemsView._hash
            h = self._hash = itemsview._hash()
        else:
            h = self._hash = hash(tuple(itemsview))
        return h



# -*- coding: utf-8 -*-

u"""
Python 2/3 compatibility helpers.

    .. py:attribute:: PY2

        True iff running on Python 2.

    .. py:attribute:: viewkeys

        viewkeys(D) → a set-like object providing a view on D's keys.

    .. py:attribute:: viewvalues

        viewvalues(D) → an object providing a view on D's values.

    .. py:attribute:: viewitems

        viewitems(D) → a set-like object providing a view on D's items.

    .. py:attribute:: iterkeys

        iterkeys(D) → an iterator over the keys of D.

    .. py:attribute:: itervalues

        itervalues(D) → an iterator over the values of D.

    .. py:attribute:: iteritems

        iteritems(D) → an iterator over the (key, value) items of D.

    .. py:attribute:: izip

        Alias for :func:`zip` on Python 3 / ``itertools.izip`` on Python 2.

    .. py:attribute:: izip_longest

        Alias for :func:`itertools.zip_longest` on Python 3 /
        ``itertools.izip_longest`` on Python 2.

"""

from operator import methodcaller
from sys import version_info

_compose = lambda f, g: lambda x: f(g(x))

PY2 = version_info[0] == 2

if PY2:  # pragma: no cover
    assert version_info[1] > 6, 'Python >= 2.7 required'
    viewkeys = methodcaller('viewkeys')
    viewvalues = methodcaller('viewvalues')
    viewitems = methodcaller('viewitems')
    iterkeys = methodcaller('iterkeys')
    itervalues = methodcaller('itervalues')
    iteritems = methodcaller('iteritems')
    from itertools import ifilter, imap, izip, izip_longest
else:  # pragma: no cover
    viewkeys = methodcaller('keys')
    viewvalues = methodcaller('values')
    viewitems = methodcaller('items')
    iterkeys = _compose(iter, viewkeys)
    itervalues = _compose(iter, viewvalues)
    iteritems = _compose(iter, viewitems)
    ifilter = filter
    imap = map
    izip = zip
    from itertools import zip_longest as izip_longest  # noqa



"""Utilities for working with one-to-one relations."""

from .compat import iteritems
from collections import Mapping
from itertools import chain


def pairs(*args, **kw):
    """
    Yield the (k, v) pairs provided, as they'd be processed if passed to *dict*.

    If a positional argument is provided,
    its pairs are yielded before those of any keyword arguments.
    The positional argument may be a mapping or sequence or pairs.

    >>> list(pairs({'a': 1}, b=2))
    [('a', 1), ('b', 2)]
    >>> list(pairs([('a', 1), ('b', 2)], b=3))
    [('a', 1), ('b', 2), ('b', 3)]

    :raises TypeError: if more than one positional arg is given.
    """
    it = None
    if args:
        arg = _arg0(args)
        if arg:
            it = iteritems(arg) if isinstance(arg, Mapping) else iter(arg)
    if kw:
        it2 = iteritems(kw)
        it = chain(it, it2) if it else it2
    return it or iter(())


def _arg0(args):
    l = len(args)
    if l != 1:
        raise TypeError('Expected at most 1 positional argument, got %d' % l)
    return args[0]


def inverted(data):
    """
    Yield the inverse items of the provided mapping or iterable.

    Works with any object that can be iterated over as a mapping or in pairs,
    or that implements its own *__inverted__* method.
    """
    inv = getattr(data, '__inverted__', None)
    return inv() if inv else _inverted(data)


def _inverted(data):
    # This is faster than `return imap(tuple, imap(reversed, pairs(data)))`:
    for (k, v) in pairs(data):
        yield (v, k)



"""Implements :class:`bidict.orderedbidict` and friends."""

from ._bidict import bidict
from ._common import BidirectionalMapping, _marker, _missing
from ._frozen import frozenbidict
from ._loose import loosebidict
from .compat import PY2, iteritems, izip
from collections import ItemsView, Mapping, OrderedDict


_PRV = 1
_NXT = 2
_END = _marker('END')


def _make_iter(reverse=False, name='__iter__', doctmpl='Like :meth:`collections.OrderedDict.%s`.'):
    def _iter(self):
        fwd = self._fwd
        end = self._end
        cur = end[_PRV if reverse else _NXT]
        while cur is not end:
            d, prv, nxt = cur
            korv = next(iter(d))
            node = fwd.get(korv)
            yield korv if node is cur else d[korv]
            cur = prv if reverse else nxt
    _iter.__name__ = name
    _iter.__doc__ = (doctmpl % name) if '%s' in doctmpl else ''
    return _iter


class OrderedBidirectionalMapping(BidirectionalMapping):
    """Base class for ordered bidirectional map types."""

    def __init__(self, *args, **kw):
        """Base impl. You probably want :func:`orderedbidict.__init__` instead."""
        self._end = []  # circular doubly-linked list of [{key: val, val: key}, prv, nxt] nodes
        self._init_end()
        self._fwd = {}  # key -> node
        self._inv = {}  # val -> node
        self._init_inv()
        if args or kw:
            self._update(True, self._on_dup_key, self._on_dup_val, self._on_dup_kv, *args, **kw)

    def _init_end(self):
        end = self._end
        end += [_END, end, end]  # sentinel node for doubly linked list

    def _init_inv(self):
        super(OrderedBidirectionalMapping, self)._init_inv()
        self.inv._end = self._end

    def copy(self):
        """Like :py:meth:`dict.copy`."""
        return self.__class__(self)

    __copy__ = copy

    def _clear(self):
        super(OrderedBidirectionalMapping, self)._clear()
        del self._end[:]
        self._init_end()

    def __getitem__(self, key):
        node = self._fwd[key]
        d = node[0]
        return d[key]

    def _pop(self, key):
        nodefwd = self._fwd.pop(key)
        d, prv, nxt = nodefwd
        val = d[key]
        nodeinv = self._inv.pop(val)
        assert nodeinv is nodefwd
        prv[_NXT] = nxt
        nxt[_PRV] = prv
        del nodefwd[:]
        return val

    def _isdupitem(self, key, val, nodeinv, nodefwd):
        """Return whether (key, val) duplicates an existing item."""
        return nodeinv is nodefwd

    def _write_item(self, key, val, isdupkey, isdupval, nodeinv, nodefwd):
        fwd = self._fwd
        inv = self._inv
        if not isdupkey and not isdupval:
            end = self._end
            lst = end[_PRV]
            node = [{key: val, val: key}, lst, end]
            end[_PRV] = lst[_NXT] = fwd[key] = inv[val] = node
            oldkey = oldval = _missing
        elif isdupkey and isdupval:
            fwdd, fwdprv, fwdnxt = nodefwd
            oldval = fwdd[key]
            invd, invprv, invnxt = nodeinv
            oldkey = invd[val]
            assert oldkey != key
            assert oldval != val
            # We have to collapse nodefwd and nodeinv into a single node.
            # Drop nodeinv so that item with same key overwritten in place.
            invprv[_NXT] = invnxt
            invnxt[_PRV] = invprv
            # Don't remove nodeinv's references to its neighbors since
            # if the update fails, we'll need them to undo this write.
            # Python's garbage collector should still be able to detect when
            # nodeinv is garbage and reclaim the memory.
            # Update fwd and inv.
            assert fwd.pop(oldkey) is nodeinv
            assert inv.pop(oldval) is nodefwd
            fwd[key] = inv[val] = nodefwd
            # Update nodefwd with new item.
            fwdd.clear()
            fwdd[key] = val
            fwdd[val] = key
        elif isdupkey:
            d = nodefwd[0]
            oldval = d[key]
            oldkey = _missing
            oldnodeinv = inv.pop(oldval)
            assert oldnodeinv is nodefwd
            inv[val] = nodefwd
        elif isdupval:
            d = nodeinv[0]
            oldkey = d[val]
            oldval = _missing
            oldnodefwd = fwd.pop(oldkey)
            assert oldnodefwd is nodeinv
            fwd[key] = nodeinv
        if isdupkey ^ isdupval:
            d.clear()
            d[key] = val
            d[val] = key
        return key, val, isdupkey, isdupval, nodeinv, nodefwd, oldkey, oldval

    def _undo_write(self, key, val, isdupkey, isdupval, nodeinv, nodefwd, oldkey, oldval):
        fwd = self._fwd
        inv = self._inv
        if not isdupkey and not isdupval:
            del self[key]
        elif isdupkey and isdupval:
            fwdd, fwdprv, fwdnxt = nodefwd
            invd, invprv, invnxt = nodeinv
            # Restore original items.
            fwdd.clear()
            fwdd[key] = oldval
            fwdd[oldval] = key
            # invd was never changed so should still have the original item.
            expectinvd = {oldkey: val, val: oldkey}
            assert invd == expectinvd
            # Undo replacing nodeinv with nodefwd.
            invprv[_NXT] = invnxt[_PRV] = nodeinv
            fwd[oldkey] = inv[val] = nodeinv
            inv[oldval] = fwd[key] = nodefwd
        elif isdupkey:
            d = nodefwd[0]
            d.clear()
            d[key] = oldval
            d[oldval] = key
            assert inv.pop(val) is nodefwd
            inv[oldval] = nodefwd
            assert fwd[key] is nodefwd
        elif isdupval:
            d = nodeinv[0]
            d.clear()
            d[oldkey] = val
            d[val] = oldkey
            assert fwd.pop(key) is nodeinv
            fwd[oldkey] = nodeinv
            assert inv[val] is nodeinv

    def __eq__(self, other):
        if not isinstance(other, Mapping) or len(self) != len(other):
            return False
        if isinstance(other, (OrderedBidirectionalMapping, OrderedDict)):
            return all(i == j for (i, j) in izip(iteritems(self), iteritems(other)))
        return all(self.get(k, _missing) == v for (k, v) in iteritems(other))

    def __repr__(self):
        inner = ', '.join('(%r, %r)' % (k, v) for (k, v) in iteritems(self))
        inner = '[%s]' % inner if inner else ''
        return '%s(%s)' % (self.__class__.__name__, inner)

    __iter__ = _make_iter()
    __reversed__ = _make_iter(reverse=True, name='__reversed__')
    if PY2:  # pragma: no cover
        def viewitems(self):
            """Like :meth:`collections.OrderedDict.viewitems`."""
            return ItemsView(self)


class orderedbidict(OrderedBidirectionalMapping, bidict):
    """Mutable, ordered bidict type."""

    def popitem(self, last=True):
        """Like :meth:`collections.OrderedDict.popitem`."""
        if not self:
            raise KeyError(self.__class__.__name__ + ' is empty')
        key = next((reversed if last else iter)(self))
        val = self._pop(key)
        return key, val

    def move_to_end(self, key, last=True):
        """Like :meth:`collections.OrderedDict.move_to_end`."""
        node = self._fwd[key]
        d, prv, nxt = node
        prv[_NXT] = nxt
        nxt[_PRV] = prv
        end = self._end
        if last:
            lst = end[_PRV]
            node[_PRV] = lst
            node[_NXT] = end
            end[_PRV] = lst[_NXT] = node
        else:
            fst = end[_NXT]
            node[_PRV] = end
            node[_NXT] = fst
            end[_NXT] = fst[_PRV] = node


class frozenorderedbidict(OrderedBidirectionalMapping, frozenbidict):
    """Immutable, hashable :class:`bidict.OrderedBidirectionalMapping` type."""


class looseorderedbidict(orderedbidict, loosebidict):
    """Mutable, ordered bidict with *OVERWRITE* duplication behaviors by default."""



"""
Implements :class:`BidirectionalMapping`, the bidirectional map base class.

Also provides related exception classes and duplication behaviors.
"""

from .compat import PY2, iteritems
from .util import pairs
from collections import Mapping


def _proxied(methodname, ivarname='_fwd', doc=None):
    """Make a func that calls methodname on the indicated instance variable."""
    def proxy(self, *args):
        ivar = getattr(self, ivarname)
        meth = getattr(ivar, methodname)
        return meth(*args)
    proxy.__name__ = methodname
    proxy.__doc__ = doc or 'Like ``dict.%s()``.' % methodname
    return proxy


class _marker(object):
    def __init__(self, id):
        self.id = id

    def __repr__(self):
        return '<%s>' % self.id


class DuplicationBehavior(_marker):
    """
    Provide RAISE, OVERWRITE, IGNORE, and ON_DUP_VAL duplication behaviors.

    .. py:attribute:: RAISE

        Raise an exception when a duplication is encountered.

    .. py:attribute:: OVERWRITE

        Overwrite an existing item when a duplication is encountered.

    .. py:attribute:: IGNORE

        Keep the existing item and ignore the new item when a duplication is
        encountered.

    .. py:attribute:: ON_DUP_VAL

        Used with *on_dup_kv* to specify that it should match whatever the
        duplication behavior of *on_dup_val* is.

    """

DuplicationBehavior.RAISE = RAISE = DuplicationBehavior('RAISE')
DuplicationBehavior.OVERWRITE = OVERWRITE = DuplicationBehavior('OVERWRITE')
DuplicationBehavior.IGNORE = IGNORE = DuplicationBehavior('IGNORE')
DuplicationBehavior.ON_DUP_VAL = ON_DUP_VAL = DuplicationBehavior('ON_DUP_VAL')
_missing = _marker('MISSING')


class BidirectionalMapping(Mapping):
    """
    Base class for all provided bidirectional map types.

    Mutable and immutable bidict types extend this class,
    which implements all the shared logic.
    Users will typically only interact with subclasses of this class.

    .. py:attribute:: inv

        The inverse bidict.

    """

    _on_dup_key = OVERWRITE
    _on_dup_val = RAISE
    _on_dup_kv = RAISE

    def __init__(self, *args, **kw):
        """Like ``dict.__init__()``, but maintaining bidirectionality."""
        self._fwd = {}  # dictionary of forward mappings
        self._inv = {}  # dictionary of inverse mappings
        self._init_inv()
        if args or kw:
            self._update(True, self._on_dup_key, self._on_dup_val, self._on_dup_kv, *args, **kw)

    def _init_inv(self):
        inv = object.__new__(self.__class__)
        inv._fwd = self._inv
        inv._inv = self._fwd
        inv.inv = self
        self.inv = inv

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self._fwd)

    def __eq__(self, other):
        return self._fwd == other

    def __ne__(self, other):
        return not self == other

    def __inverted__(self):
        """Get an iterator over the items in :attr:`self.inv <inv>`."""
        return iteritems(self.inv)

    def _pop(self, key):
        val = self._fwd.pop(key)
        del self._inv[val]
        return val

    def _clear(self):
        self._fwd.clear()
        self._inv.clear()

    def _put(self, key, val, on_dup_key, on_dup_val, on_dup_kv):
        result = self._dedup_item(key, val, on_dup_key, on_dup_val, on_dup_kv)
        if result:
            self._write_item(key, val, *result)

    def _dedup_item(self, key, val, on_dup_key, on_dup_val, on_dup_kv):
        """
        Check *key* and *val* for any duplication in self.

        Handle any duplication as per the given duplication behaviors.

        (key, val) already present is construed as a no-op, not a duplication.

        If duplication is found and the corresponding duplication behavior is
        *RAISE*, raise the appropriate error.

        If duplication is found and the corresponding duplication behavior is
        *IGNORE*, return *None*.

        If duplication is found and the corresponding duplication behavior is
        *OVERWRITE*, or if no duplication is found, return the dedup result
        *(isdupkey, isdupval, invbyval, fwdbykey)*.
        """
        fwd = self._fwd
        inv = self._inv
        fwdbykey = fwd.get(key, _missing)
        invbyval = inv.get(val, _missing)
        isdupkey = fwdbykey is not _missing
        isdupval = invbyval is not _missing
        if isdupkey and isdupval:
            if self._isdupitem(key, val, invbyval, fwdbykey):
                # (key, val) duplicates an existing item -> no-op.
                return
            # key and val each duplicate a different existing item.
            if on_dup_kv is RAISE:
                raise KeyAndValueDuplicationError(key, val)
            elif on_dup_kv is IGNORE:
                return
            # else on_dup_kv is OVERWRITE. Fall through to return on last line.
        elif isdupkey:
            if on_dup_key is RAISE:
                raise KeyDuplicationError(key)
            elif on_dup_key is IGNORE:
                return
            # else on_dup_key is OVERWRITE. Fall through to return on last line.
        elif isdupval:
            if on_dup_val is RAISE:
                raise ValueDuplicationError(val)
            elif on_dup_val is IGNORE:
                return
            # else on_dup_val is OVERWRITE. Fall through to return on last line.
        # else neither isdupkey nor isdupval.
        return isdupkey, isdupval, invbyval, fwdbykey

    def _isdupitem(self, key, val, oldkey, oldval):
        dup = oldkey == key
        assert dup == (oldval == val)
        return dup

    def _write_item(self, key, val, isdupkey, isdupval, oldkey, oldval):
        self._fwd[key] = val
        self._inv[val] = key
        if isdupkey:
            del self._inv[oldval]
        if isdupval:
            del self._fwd[oldkey]
        return key, val, isdupkey, isdupval, oldkey, oldval

    def _update(self, init, on_dup_key, on_dup_val, on_dup_kv, *args, **kw):
        if not args and not kw:
            return
        if on_dup_kv is ON_DUP_VAL:
            on_dup_kv = on_dup_val
        rollbackonfail = not init or RAISE in (on_dup_key, on_dup_val, on_dup_kv)
        if rollbackonfail:
            return self._update_rbf(on_dup_key, on_dup_val, on_dup_kv, *args, **kw)
        _put = self._put
        for (key, val) in pairs(*args, **kw):
            _put(key, val, on_dup_key, on_dup_val, on_dup_kv)

    def _update_rbf(self, on_dup_key, on_dup_val, on_dup_kv, *args, **kw):
        """Update, rolling back on failure."""
        exc = None
        writes = []
        appendwrite = writes.append
        dedup_item = self._dedup_item
        write_item = self._write_item
        for (key, val) in pairs(*args, **kw):
            try:
                dedup_result = dedup_item(key, val, on_dup_key, on_dup_val, on_dup_kv)
            except DuplicationError as e:
                exc = e
                break
            if dedup_result:
                write_result = write_item(key, val, *dedup_result)
                appendwrite(write_result)
        if exc:
            undo_write = self._undo_write
            for write in reversed(writes):
                undo_write(*write)
            raise exc

    def _undo_write(self, key, val, isdupkey, isdupval, oldkey, oldval):
        fwd = self._fwd
        inv = self._inv
        if not isdupkey and not isdupval:
            self._pop(key)
            return
        if isdupkey:
            fwd[key] = oldval
            inv[oldval] = key
            if not isdupval:
                del inv[val]
        if isdupval:
            inv[val] = oldkey
            fwd[oldkey] = val
            if not isdupkey:
                del fwd[key]

    def copy(self):
        """Like :py:meth:`dict.copy`."""
        copy = object.__new__(self.__class__)
        copy._fwd = self._fwd.copy()
        copy._inv = self._inv.copy()
        cinv = object.__new__(self.__class__)
        cinv._fwd = copy._inv
        cinv._inv = copy._fwd
        cinv.inv = copy
        copy.inv = cinv
        return copy

    __copy__ = copy
    __len__ = _proxied('__len__')
    __iter__ = _proxied('__iter__')
    __getitem__ = _proxied('__getitem__')
    values = _proxied('keys', ivarname='inv')
    values.__doc__ = \
        "B.values() -> a set-like object providing a view on B's values.\n\n" \
        'Note that because values of a BidirectionalMapping are also keys\n' \
        'of its inverse, this returns a *KeysView* object rather than a\n' \
        '*ValuesView* object, conferring set-like benefits.'
    if PY2:  # pragma: no cover
        viewkeys = _proxied('viewkeys')
        viewitems = _proxied('viewitems')
        itervalues = _proxied('iterkeys', ivarname='inv',
                              doc=dict.itervalues.__doc__)
        viewvalues = _proxied('viewkeys', ivarname='inv',
                              doc=values.__doc__.replace('values()', 'viewvalues()'))
        values.__doc__ = 'Like ``dict.values()``.'


class BidictException(Exception):
    """Base class for bidict exceptions."""


class DuplicationError(BidictException):
    """Base class for exceptions raised when uniqueness is violated."""


class KeyDuplicationError(DuplicationError):
    """Raised when a given key is not unique."""


class ValueDuplicationError(DuplicationError):
    """Raised when a given value is not unique."""


class KeyAndValueDuplicationError(KeyDuplicationError, ValueDuplicationError):
    """
    Raised when a given item's key and value are not unique.

    That is, its key duplicates that of another item,
    and its value duplicates that of a different other item.
    """



"""Implements :class:`bidict.loosebidict`."""

from ._common import OVERWRITE
from ._bidict import bidict


class loosebidict(bidict):
    """Mutable bidict with *OVERWRITE* duplication behaviors by default."""

    _on_dup_val = OVERWRITE
    _on_dup_kv = OVERWRITE



"""Implements :class:`bidict.bidict`, the mutable bidirectional map type."""

from ._common import BidirectionalMapping, OVERWRITE, RAISE, ON_DUP_VAL
from collections import MutableMapping


class bidict(BidirectionalMapping, MutableMapping):
    """Mutable bidirectional map type."""

    def __delitem__(self, key):
        """Like :py:meth:`dict.__delitem__`, maintaining bidirectionality."""
        self._pop(key)

    def __setitem__(self, key, val):
        """
        Set the value for *key* to *val*.

        If *key* is already associated with *val*, this is a no-op.

        If *key* is already associated with a different value,
        the old value will be replaced with *val*,
        as with :py:meth:`dict.__setitem__`.

        If *val* is already associated with a different key,
        an exception is raised
        to protect against accidental removal of the key
        that's currently associated with *val*.

        Use :attr:`put` instead if you want to specify different behavior in
        the case that the provided key or value duplicates an existing one.
        Or use :attr:`forceput` to unconditionally associate *key* with *val*,
        replacing any existing items as necessary to preserve uniqueness.

        :raises bidict.ValueDuplicationError: if *val* duplicates that of an
            existing item.

        :raises bidict.KeyAndValueDuplicationError: if *key* duplicates the key of an
            existing item and *val* duplicates the value of a different
            existing item.
        """
        self._put(key, val, self._on_dup_key, self._on_dup_val, self._on_dup_kv)

    def put(self, key, val, on_dup_key=RAISE, on_dup_val=RAISE, on_dup_kv=ON_DUP_VAL):
        """
        Associate *key* with *val* with the specified duplication behaviors.

        For example, if all given duplication behaviors are
        :attr:`DuplicationBehavior.RAISE <bidict.DuplicationBehavior.RAISE>`,
        then *key* will be associated with *val* if and only if
        *key* is not already associated with an existing value and
        *val* is not already associated with an existing key,
        otherwise an exception will be raised.

        If *key* is already associated with *val*, this is a no-op.

        :raises bidict.KeyDuplicationError: if attempting to insert an item
            whose key duplicates an existing item's, and *on_dup_key* is
            :attr:`RAISE <bidict.DuplicationBehavior.RAISE>`.

        :raises bidict.ValueDuplicationError: if attempting to insert an item
            whose value duplicates an existing item's, and *on_dup_val* is
            :attr:`RAISE <bidict.DuplicationBehavior.RAISE>`.

        :raises bidict.KeyAndValueDuplicationError: if attempting to insert an
            item whose key duplicates one existing item's, and whose value
            duplicates another existing item's, and *on_dup_kv* is
            :attr:`RAISE <bidict.DuplicationBehavior.RAISE>`.
        """
        self._put(key, val, on_dup_key, on_dup_val, on_dup_kv)

    def forceput(self, key, val):
        """
        Associate *key* with *val* unconditionally.

        Replace any existing mappings containing key *key* or value *val*
        as necessary to preserve uniqueness.
        """
        self._put(key, val, OVERWRITE, OVERWRITE, OVERWRITE)

    def clear(self):
        """Remove all items."""
        self._clear()

    def pop(self, key, *args):
        """Like :py:meth:`dict.pop`, maintaining bidirectionality."""
        l = len(args) + 1
        if l > 2:
            raise TypeError('pop expected at most 2 arguments, got %d' % l)
        try:
            return self._pop(key)
        except KeyError:
            if args:
                return args[0]  # default
            raise

    def popitem(self):
        """Like :py:meth:`dict.popitem`, maintaining bidirectionality."""
        if not self._fwd:
            raise KeyError('popitem(): %s is empty' % self.__class__.__name__)
        key, val = self._fwd.popitem()
        del self._inv[val]
        return key, val

    def setdefault(self, key, default=None):
        """Like :py:meth:`dict.setdefault`, maintaining bidirectionality."""
        if key not in self:
            self[key] = default
        return self[key]

    def update(self, *args, **kw):
        """
        Like :attr:`putall` with default duplication behaviors.

        In particular, for :class:`bidict.bidict`,
        *on_dup_key=OVERWRITE*, *on_dup_val=RAISE*, and *on_dup_kv=RAISE*.

        For :class:`bidict.loosebidict`,
        *on_dup_key=OVERWRITE*, *on_dup_val=OVERWRITE*, and *on_dup_kv=OVERWRITE*.
        """
        self._update(False, self._on_dup_key, self._on_dup_val, self._on_dup_kv, *args, **kw)

    def forceupdate(self, *args, **kw):
        """Like a bulk :attr:`forceput`."""
        self._update(False, OVERWRITE, OVERWRITE, OVERWRITE, *args, **kw)

    def putall(self, items, on_dup_key=RAISE, on_dup_val=RAISE, on_dup_kv=ON_DUP_VAL):
        """
        Like a bulk :attr:`put`.

        If one of the given items causes an exception to be raised,
        none of the items is inserted.
        """
        self._update(False, on_dup_key, on_dup_val, on_dup_kv, items)



# -*- coding: utf-8 -*-

"""
Efficient, Pythonic bidirectional map implementation and related functionality.

See https://bidict.readthedocs.io for comprehensive documentation.

.. :copyright: (c) 2016 Joshua Bronson.
.. :license: ISCL. See LICENSE for details.

"""

from ._common import (BidirectionalMapping, BidictException,
                      DuplicationBehavior, IGNORE, OVERWRITE, RAISE,
                      DuplicationError, KeyDuplicationError, ValueDuplicationError,
                      KeyAndValueDuplicationError)
from ._bidict import bidict
from ._loose import loosebidict
from ._frozen import frozenbidict
from ._named import namedbidict
from ._ordered import (OrderedBidirectionalMapping,
                       orderedbidict, frozenorderedbidict, looseorderedbidict)
from .util import pairs, inverted

__all__ = (
    'BidirectionalMapping',
    'BidictException',
    'DuplicationBehavior',
    'IGNORE',
    'OVERWRITE',
    'RAISE',
    'DuplicationError',
    'KeyDuplicationError',
    'ValueDuplicationError',
    'KeyAndValueDuplicationError',
    'bidict',
    'loosebidict',
    'frozenbidict',
    'namedbidict',
    'OrderedBidirectionalMapping',
    'orderedbidict',
    'frozenorderedbidict',
    'looseorderedbidict',
    'pairs',
    'inverted',
)

try:
    from pkg_resources import resource_string
    __version__ = resource_string(__name__, 'VERSION').decode('ascii').strip()
except Exception as e:  # pragma: no cover
    from warnings import warn
    warn('Failed to read/set version: %r' % e)



from io import open
from setuptools import setup
from warnings import warn

try:
    with open('bidict/VERSION', encoding='utf8') as f:
        version = f.read().strip()
except Exception as e:
    version = '0.0.0'
    warn('Could not open bidict/VERSION, using bogus version (%s): %r' %
         (version, e))
try:
    with open('README.rst', encoding='utf8') as f:
        long_description = f.read()
except Exception as e:
    long_description = 'See https://bidict.readthedocs.org'
    warn('Could not open README.rst, using provisional long_description '
         '(%r): %r' % (long_description, e))

tests_require = [
    'coverage==4.1',
    'flake8==2.5.4',
    'hypothesis==3.4.0',
    'hypothesis-pytest==0.19.0',
    'py==1.4.31',
    'pydocstyle==1.0.0',
    'pytest==2.9.2',
    'pytest-benchmark==3.0.0',
    'pytest-cov==2.2.1',
    'Sphinx==1.4.4',
]

setup(
    name='bidict',
    version=version,
    author='Joshua Bronson',
    author_email='jab@math.brown.edu',
    description='Efficient, Pythonic bidirectional map implementation and related functionality',
    long_description=long_description,
    keywords='dict, dictionary, mapping, bidirectional, bijection, bijective, injective, two-way, 2-way, double, inverse, reverse',
    url='https://github.com/jab/bidict',
    license='ISCL',
    packages=['bidict'],
    package_data=dict(bidict=['VERSION']),
    zip_safe=True,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ],
    tests_require=tests_require,
    extras_require=dict(
        test=tests_require,
        dev=tests_require + ['pre-commit==0.7.6', 'tox==2.3.1'],
    ),
)



# -*- coding: utf-8 -*-

"""Implements :class:`bidict.namedbidict`."""

from ._bidict import bidict
from re import compile as compile_re


_LEGALNAMEPAT = '^[a-zA-Z][a-zA-Z0-9_]*$'
_LEGALNAMERE = compile_re(_LEGALNAMEPAT)


def namedbidict(typename, keyname, valname, base_type=bidict):
    """
    Create a bidict type with custom accessors.

    Analagous to :func:`collections.namedtuple`.
    """
    for name in typename, keyname, valname:
        if not _LEGALNAMERE.match(name):
            raise ValueError('"%s" does not match pattern %s' %
                             (name, _LEGALNAMEPAT))

    getfwd = lambda self: self
    getfwd.__name__ = valname + '_for'
    getfwd.__doc__ = '%s forward %s: %s → %s' % (typename, base_type.__name__, keyname, valname)

    getinv = lambda self: self.inv
    getinv.__name__ = keyname + '_for'
    getinv.__doc__ = '%s inverse %s: %s → %s' % (typename, base_type.__name__, valname, keyname)

    __reduce__ = lambda self: (_make_empty, (typename, keyname, valname), self.__dict__)
    __reduce__.__name__ = '__reduce__'
    __reduce__.__doc__ = 'helper for pickle'

    __dict__ = {
        getfwd.__name__: property(getfwd),
        getinv.__name__: property(getinv),
        '__reduce__': __reduce__,
    }
    return type(typename, (base_type,), __dict__)


def _make_empty(typename, keyname, valname):
    """
    Create an empty instance of a custom bidict.

    Used to make :func:`bidict.namedbidict` instances picklable.
    """
    named = namedbidict(typename, keyname, valname)
    return named()
